/////////////////////////////////////////////////////////////////////////////
// MOUNT.EXE - (c)2002 Alin Constantin
//
// Mount is an application that displays the volumes available on the local machine,
// allowing to mount and unmount them as folders or drives.
/////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"

#define _WIN32_WINNT 0x0500

#include <windows.h>
#include <stdio.h>
#include <direct.h>
#include <stdlib.h>
#include <errno.h>
#include <io.h>

// NOTE: The string buffer needs to be large.
// Some functions like FindFirstVolumeMountPoint, GetVolumePathNamesForVolumeName would fail with ERROR_MORE_DATA
// when the buffer cannot hold the mount point. We need to have this buffer size at least _MAX_PATH
#define BUFSIZE 1024				// ample space for our string buffer
#define FILESYSTEMNAMEBUFSIZE 100	// probably more than we need for file system type

#define NUMBER_OF(x) (sizeof(x)/sizeof(*x))

typedef WINBASEAPI HANDLE ( WINAPI Type_FindFirstVolume)
(
    LPTSTR lpszVolumeName,
    DWORD cchBufferLength
);

typedef WINBASEAPI BOOL ( WINAPI Type_FindNextVolume)
(
    HANDLE hFindVolume,
    LPTSTR lpszVolumeName,
    DWORD cchBufferLength
);

typedef WINBASEAPI BOOL ( WINAPI Type_FindVolumeClose)
(
    HANDLE hFindVolume
);

typedef WINBASEAPI HANDLE ( WINAPI Type_FindFirstVolumeMountPoint)
(
    LPCTSTR lpszRootPathName,
    LPTSTR lpszVolumeMountPoint,
    DWORD cchBufferLength
);

typedef WINBASEAPI BOOL ( WINAPI Type_FindNextVolumeMountPoint)
(
    HANDLE hFindVolumeMountPoint,
    LPTSTR lpszVolumeMountPoint,
    DWORD cchBufferLength
);

typedef WINBASEAPI BOOL ( WINAPI Type_FindVolumeMountPointClose)
(
    HANDLE hFindVolumeMountPoint
);

typedef WINBASEAPI BOOL ( WINAPI Type_GetVolumePathNamesForVolumeName)
(
  LPCTSTR lpszVolumeName,
  LPTSTR lpszVolumePathNames,
  DWORD cchBufferLength,
  PDWORD lpcchReturnLength
);

typedef WINBASEAPI BOOL ( WINAPI Type_GetVolumeNameForVolumeMountPoint)
(
    LPCTSTR lpszVolumeMountPoint,
    LPTSTR lpszVolumeName,
    DWORD cchBufferLength
);

typedef WINBASEAPI BOOL ( WINAPI Type_GetVolumeInformation)
(
    LPCTSTR lpRootPathName,
    LPTSTR lpVolumeNameBuffer,
    DWORD nVolumeNameSize,
    LPDWORD lpVolumeSerialNumber,
    LPDWORD lpMaximumComponentLength,
    LPDWORD lpFileSystemFlags,
    LPTSTR lpFileSystemNameBuffer,
    DWORD nFileSystemNameSize
);

typedef WINBASEAPI BOOL ( WINAPI Type_SetVolumeMountPoint)
(
    LPCTSTR lpszVolumeMountPoint,
    LPCTSTR lpszVolumeName
);

typedef WINBASEAPI BOOL ( WINAPI Type_DeleteVolumeMountPoint)
(
    LPCTSTR lpszVolumeMountPoint
);

// The functions imported from Kernel32.dll
Type_FindFirstVolume * g_lpfnFindFirstVolume = NULL;
Type_FindNextVolume * g_lpfnFindNextVolume = NULL;
Type_FindVolumeClose * g_lpfnFindVolumeClose = NULL;

Type_FindFirstVolumeMountPoint * g_lpfnFindFirstVolumeMountPoint = NULL;
Type_FindNextVolumeMountPoint * g_lpfnFindNextVolumeMountPoint = NULL;
Type_FindVolumeMountPointClose * g_lpfnFindVolumeMountPointClose = NULL;

Type_GetVolumeInformation * g_lpfnGetVolumeInformation = NULL;
Type_GetVolumeNameForVolumeMountPoint * g_lpfnGetVolumeNameForVolumeMountPoint = NULL;
Type_GetVolumePathNamesForVolumeName * g_lpfnGetVolumePathNamesForVolumeName = NULL;

Type_SetVolumeMountPoint * g_lpfnSetVolumeMountPoint = NULL;
Type_DeleteVolumeMountPoint *  g_lpfnDeleteVolumeMountPoint = NULL;

// The map of drives letters already assigned to volumes
TCHAR * g_szDriveAssignement['Z'-'A'+1];

/////////////////////////////////////////////////////////////////////////////
// vWriteHelpHeader()
//
// Displays the header of the help message
/////////////////////////////////////////////////////////////////////////////
void
vWriteHelpHeader (void)
{
	_tprintf( _T("\n") );
	_tprintf( _T("========================================================================\n"));
	_tprintf( _T("MOUNT.EXE v1.1 - (c)2002 Alin Constantin\n") );
	_tprintf( _T("========================================================================\n"));
	_tprintf( _T("\n") );
}

/////////////////////////////////////////////////////////////////////////////
// vMapFunctionsAddresses()
//
// Get the addresses of supported functions
/////////////////////////////////////////////////////////////////////////////
void
vMapFunctionsAddresses (void)
{
	// Check whether GetVolumePathNamesForVolumeName is supported 
	HMODULE hModule = ::LoadLibrary( _T("Kernel32.dll") );

	g_lpfnFindFirstVolume = reinterpret_cast<Type_FindFirstVolume *>(
		::GetProcAddress (hModule, 
		#ifdef UNICODE
		"FindFirstVolumeW"
		#else
		"FindFirstVolumeA"
		#endif
		) );

	g_lpfnFindNextVolume = reinterpret_cast<Type_FindNextVolume *>(
		::GetProcAddress (hModule, 
		#ifdef UNICODE
		"FindNextVolumeW"
		#else
		"FindNextVolumeA"
		#endif
		) );

	g_lpfnFindVolumeClose = reinterpret_cast<Type_FindVolumeClose *>(
		::GetProcAddress (hModule, 	"FindVolumeClose") );

	g_lpfnFindFirstVolumeMountPoint = reinterpret_cast<Type_FindFirstVolumeMountPoint *>(
		::GetProcAddress (hModule, 
		#ifdef UNICODE
		"FindFirstVolumeMountPointW"
		#else
		"FindFirstVolumeMountPointA"
		#endif
		) );

	g_lpfnFindNextVolumeMountPoint = reinterpret_cast<Type_FindNextVolumeMountPoint *>(
		::GetProcAddress (hModule, 
		#ifdef UNICODE
		"FindNextVolumeMountPointW"
		#else
		"FindNextVolumeMountPointA"
		#endif
		) );

	g_lpfnFindVolumeMountPointClose = reinterpret_cast<Type_FindVolumeMountPointClose *>(
		::GetProcAddress (hModule, "FindVolumeMountPointClose") );

	g_lpfnGetVolumeInformation = reinterpret_cast<Type_GetVolumeInformation *>(
		::GetProcAddress (hModule, 
		#ifdef UNICODE
		"GetVolumeInformationW"
		#else
		"GetVolumeInformationA"
		#endif
		) );

	g_lpfnGetVolumeNameForVolumeMountPoint = reinterpret_cast<Type_GetVolumeNameForVolumeMountPoint *>(
		::GetProcAddress (hModule, 
		#ifdef UNICODE
		"GetVolumeNameForVolumeMountPointW"
		#else
		"GetVolumeNameForVolumeMountPointA"
		#endif
		) );

	g_lpfnGetVolumePathNamesForVolumeName = reinterpret_cast<Type_GetVolumePathNamesForVolumeName *>(
		::GetProcAddress (hModule, 
		#ifdef UNICODE
		"GetVolumePathNamesForVolumeNameW"
		#else
		"GetVolumePathNamesForVolumeNameA"
		#endif
		) );

	g_lpfnSetVolumeMountPoint = reinterpret_cast<Type_SetVolumeMountPoint *>(
		::GetProcAddress (hModule, 
		#ifdef UNICODE
		"SetVolumeMountPointW"
		#else
		"SetVolumeMountPointA"
		#endif
		) );

	g_lpfnDeleteVolumeMountPoint = reinterpret_cast<Type_DeleteVolumeMountPoint *>(
		::GetProcAddress (hModule, 
		#ifdef UNICODE
		"DeleteVolumeMountPointW"
		#else
		"DeleteVolumeMountPointA"
		#endif
		) );

	::FreeLibrary(hModule);

	// Verify if we are running on Windows 2000+
	// On Windows 2000 GetVolumePathNamesForVolumeName is not implemented, but we can workaround it
	if (!g_lpfnFindFirstVolume || !g_lpfnFindNextVolume ||	!g_lpfnFindVolumeClose ||
		!g_lpfnFindFirstVolumeMountPoint || !g_lpfnFindNextVolumeMountPoint || !g_lpfnFindVolumeMountPointClose ||
		!g_lpfnGetVolumeInformation || !g_lpfnGetVolumeNameForVolumeMountPoint ||
		!g_lpfnSetVolumeMountPoint || !g_lpfnDeleteVolumeMountPoint )
	{
		vWriteHelpHeader();
		_tprintf( _T("Sorry, but this program does not run on older systems like\n"));
		_tprintf( _T("Windows 95, Windows 98, Windows Millenium or Windows NT4.\n") );
		_tprintf( _T("\n"));
		_tprintf( _T("Mount points requires NTFS 5.0 file system types because they are implemented\n"));
		_tprintf( _T("by Windows using reparse points. This is a feature of NTFS 5.0 file system and\n"));
		_tprintf( _T("is supported only by Windows 2000 and later versions."));
		exit (-1);
	}
}

/////////////////////////////////////////////////////////////////////////////
// vWriteSyntax()
//
// Displays the syntax of the program
/////////////////////////////////////////////////////////////////////////////
void
vWriteSyntax (void)
{
	vWriteHelpHeader();
	_tprintf( _T("Syntax:\n") );
	_tprintf( _T("\n") );
	_tprintf( _T("  MOUNT <volume> <mount_point> - Mount a volume at the specified mount point.\n"));
	_tprintf( _T("\tE.g.: mount F:\\ C:\\Mnt\\DriveF\\\n") );
	_tprintf( _T("\t      mount \\\\?\\Volume{5a3600b1-e826-11d5-966f-806d6172696f}\\ X:\\\n") );
	_tprintf( _T("\n") );
	_tprintf( _T("  MOUNT /D <mount_point> or\n"));
	_tprintf( _T("  MOUNT /U <mount_point> - Unmount a volume from the specified point.\n"));
	_tprintf( _T("\tE.g.: mount /D X:\\\n") );
	_tprintf( _T("\t      mount /U C:\\Mnt\\DriveF\\\n") );
	_tprintf( _T("\n") );
	_tprintf( _T("  MOUNT /L - List available volumes and current mount points.\n"));
}

/////////////////////////////////////////////////////////////////////////////
// iDeallocateDriveAssignment()
//
// Deallocate the memory used for drives assignment
/////////////////////////////////////////////////////////////////////////////
int
iDeallocateDriveAssignment
(
)
{
	// If there were some drive mapping stored, deallocate the memory
	for (int iDriveAssigned = 0; iDriveAssigned < NUMBER_OF(g_szDriveAssignement); iDriveAssigned++)
	{
		if (g_szDriveAssignement[iDriveAssigned] != NULL)
		{
			delete [] g_szDriveAssignement[iDriveAssigned];
		}
	}

	return 0;
}

/////////////////////////////////////////////////////////////////////////////
// MAIN - the main function of the application
//
// Returns:  1 - success
//			 0 - failure
//			-1 - incorrect syntax
/////////////////////////////////////////////////////////////////////////////
int _tmain(int argc, _TCHAR* argv[])
{
	BOOL bFlag = TRUE;
	TCHAR szVolumeName[BUFSIZE];  // buffer for volume name

	memset(g_szDriveAssignement, 0, sizeof(g_szDriveAssignement));
	_onexit(iDeallocateDriveAssignment);

	vMapFunctionsAddresses();

	// Check arguments
	if( argc <= 1 ) 
	{
		vWriteSyntax( );
		return( -1 );
	}

	// Set the error mode such that system will not ask to insert a diskette in a disk drive or a CD in CDROM drive
	::SetErrorMode (SEM_FAILCRITICALERRORS);
	TCHAR szDriveRoot[] = _T("C:\\");
	for (int iDriveAssigned = 0; iDriveAssigned < NUMBER_OF(g_szDriveAssignement); iDriveAssigned++)
	{
		*szDriveRoot = iDriveAssigned + 'A';
		if ((*g_lpfnGetVolumeNameForVolumeMountPoint)(szDriveRoot, szVolumeName, NUMBER_OF(szVolumeName)))
		{
			g_szDriveAssignement[iDriveAssigned] = new TCHAR [_tcslen(szVolumeName) + 1];
			_tcscpy( g_szDriveAssignement[iDriveAssigned], szVolumeName );
		}
	}

	// Check for a list operation
	if ( (*argv[1]=='/' || *argv[1]=='-') && _totupper(argv[1][1])=='L')
	{
		// List command require one argument only
		if( argc != 2 ) 
		{
			vWriteSyntax( );
			return( -1 );
		}

		_tprintf( _T("Volumes:\n"));

		// On Windows 2000 drives needs to be enumerated to get the association with their volume name
		// because there is no easy way to get the letters from the volume name
		if (!g_lpfnGetVolumePathNamesForVolumeName)
		{
			// Set the error mode such that system will not ask to insert a diskette in a disk drive or a CD in CDROM drive
			::SetErrorMode (SEM_FAILCRITICALERRORS);
			TCHAR szDriveRoot[] = _T("C:\\");
			for (int iDriveAssigned = 0; iDriveAssigned < NUMBER_OF(g_szDriveAssignement); iDriveAssigned++)
			{
				if (g_szDriveAssignement[iDriveAssigned])
				{
					*szDriveRoot = iDriveAssigned + 'A';
					_tprintf( _T("%s - Mounted as %s\n"), g_szDriveAssignement[iDriveAssigned], szDriveRoot);
				}
			}
		}

		HANDLE hVolume = (*g_lpfnFindFirstVolume)(szVolumeName, NUMBER_OF(szVolumeName));
		if (hVolume == INVALID_HANDLE_VALUE)
		{
			_tprintf( _T("Cannot enumerate volumes\n"));
		}
		else
		{
			do 
			{
				if (g_lpfnGetVolumePathNamesForVolumeName != NULL)
				{
					// The function is supported (Windows XP/Windows .NET), let's use the newer function

					TCHAR szVolumePaths[100];
					DWORD dwRequiredSize = NUMBER_OF(szVolumePaths);
					TCHAR * pszDynamicVolPaths = NULL;
					TCHAR * pszVolumePaths = szVolumePaths;
					DWORD dwLastError = ERROR_SUCCESS;

					if (! (*g_lpfnGetVolumePathNamesForVolumeName)(szVolumeName, szVolumePaths, NUMBER_OF(szVolumePaths), &dwRequiredSize)) 
					{
						// According with MSDN, the dwRequiredSize should contain the required buffer size if the error is ERROR_MORE_DATA
						// According with my observations, this is returned correctly only when the function is successfull :-(
						// Since we cannot use directly that value, call the function doubling each time the buffer size
						dwLastError = ::GetLastError();
						while (dwLastError==ERROR_MORE_DATA)
						{
							// prevent overflowing the dword value
							if (2*dwRequiredSize < dwRequiredSize)
							{
								break;
							}

							dwRequiredSize = 2*dwRequiredSize;
							if (pszDynamicVolPaths != NULL)
							{
								delete pszDynamicVolPaths;
							}
							
							pszDynamicVolPaths = new TCHAR[dwRequiredSize];
							pszVolumePaths = pszDynamicVolPaths;
							
							dwLastError = ERROR_SUCCESS;
							if (!(*g_lpfnGetVolumePathNamesForVolumeName)(szVolumeName, pszVolumePaths, dwRequiredSize, &dwRequiredSize))
							{
								dwLastError = ::GetLastError();
							}
						}
					}

					if (dwLastError != ERROR_SUCCESS)
					{
						_tprintf( _T("%s - Cannot get mount info\n"), szVolumeName);
					}
					else
					{
						if (!pszVolumePaths || !*pszVolumePaths)
						{
							_tprintf( _T("%s - Not mounted\n"), szVolumeName);
						}
						else
						{
							while (*pszVolumePaths)
							{
								_tprintf( _T("%s - Mounted as %s\n"), szVolumeName, pszVolumePaths);
								pszVolumePaths += _tcslen(pszVolumePaths) + 1;
							}
						}
					}

					if (pszDynamicVolPaths != NULL)
					{
						delete pszDynamicVolPaths;
					}
				}
				else
				{
					// Windows 2000, let's try alternate method of enumerating the volumes

					// find the drive letter of this volume, we'll need it later
					TCHAR szDriveRoot[] = _T("C:\\");
					for (int iDriveAssigned = 0; iDriveAssigned < NUMBER_OF(g_szDriveAssignement); iDriveAssigned++)
					{
						if ( g_szDriveAssignement[iDriveAssigned] != NULL && !_tcsicmp(szVolumeName, g_szDriveAssignement[iDriveAssigned]) )
						{
							*szDriveRoot = iDriveAssigned + 'A';
							break;
						}
					}

					if (iDriveAssigned < NUMBER_OF(g_szDriveAssignement))
					{
						DWORD dwSysFlags;   // flags that describe the file system
						TCHAR szVolumeFileSystemNameBuffer[FILESYSTEMNAMEBUFSIZE]; // file system name (FAT, NTFS, etc)

						// Set the error mode such that system will not ask to insert a diskette in a disk drive or a CD in CDROM drive
						::SetErrorMode (SEM_FAILCRITICALERRORS);
						if ((*g_lpfnGetVolumeInformation)( szVolumeName, NULL, 0, NULL, NULL, &dwSysFlags, szVolumeFileSystemNameBuffer, 
									NUMBER_OF(szVolumeFileSystemNameBuffer)) )
						{
							// Detect support for reparse points, and therefore for volume 
							// mount points, which are implemented using reparse points. 
							if (dwSysFlags & FILE_SUPPORTS_REPARSE_POINTS) 
							{
								// Start processing mount points on this volume. 
								TCHAR szMountPoint[BUFSIZE];
								HANDLE hMountPoint = (*g_lpfnFindFirstVolumeMountPoint)( szVolumeName, 
										szMountPoint, NUMBER_OF(szMountPoint) );

								if (hMountPoint != INVALID_HANDLE_VALUE)
								{
									do 
									{
										// Build a unique path to the mount point
										TCHAR szMountPath[BUFSIZE];
										_tcsncpy (szMountPath, szDriveRoot, NUMBER_OF(szMountPath));           
										_tcsncat (szMountPath, szMountPoint, NUMBER_OF(szMountPath) - _tcslen(szMountPath));

										TCHAR szMountedVolumeName[BUFSIZE];   // target of mount at mount point
										if ((*g_lpfnGetVolumeNameForVolumeMountPoint)( szMountPath, szMountedVolumeName, NUMBER_OF(szMountedVolumeName)))
										{
											_tprintf( _T("%s - Mounted as %s\n"), szMountedVolumeName, szMountPath);
										}

									} while ( (*g_lpfnFindNextVolumeMountPoint)(hMountPoint, szVolumeName, NUMBER_OF(szVolumeName)) );
										
									(*g_lpfnFindVolumeMountPointClose)(hMountPoint);
								}
							}
						}
					}
					else
					{
						_tprintf( _T("%s - Not mounted\n"), szVolumeName);
					}
				} 
		
			} while ( (*g_lpfnFindNextVolume)(hVolume, szVolumeName, NUMBER_OF(szVolumeName)) );
			
			(*g_lpfnFindVolumeClose)(hVolume);
		}
	}
	else
	{
		// Both mount and unmount commands require 2 arguments.
		if( argc != 3 ) 
		{
			vWriteSyntax( );
			return( -1 );
		}

		bool fMountingVolumne = true;
		if ( (*argv[1]=='/' || *argv[1]=='-') && (_totupper(argv[1][1])=='U' || _totupper(argv[1][1])=='D') )
		{
			fMountingVolumne = false;
		}

		TCHAR szMountPoint[BUFSIZE];  // buffer for the mount path
		_tcsncpy( szMountPoint, argv[2], NUMBER_OF(szMountPoint) );
		// make sure the string is NULL terminated
		szMountPoint[NUMBER_OF(szMountPoint)-1] = '\0';

		// Make sure the mount point is terminated in a slash
		size_t iPathLength = _tcslen(szMountPoint);
		if (iPathLength && iPathLength < NUMBER_OF(szMountPoint) && szMountPoint[iPathLength-1] != '\\')
		{
			_tcscat(szMountPoint, _T("\\"));
		}

		if (fMountingVolumne)
		{
			// For user's convenience, we will accept both volume names and drive letters for the mount command.
			if ( _tcslen(argv[1])>=2 && argv[1][1]==':')
			{
				// Drive letter was speciifed, let's identify the volume name
				TCHAR szDriveLetter[_MAX_DRIVE+1];
				_tcsncpy(szDriveLetter, argv[1], _MAX_DRIVE);
				szDriveLetter[_MAX_DRIVE-1]='\\';
				szDriveLetter[_MAX_DRIVE]='\0';

				bFlag = (*g_lpfnGetVolumeNameForVolumeMountPoint)(
					szDriveLetter, // input volume mount point or directory
					szVolumeName, // output volume name buffer
					NUMBER_OF(szVolumeName) // size of volume name buffer
					);

				if (bFlag != TRUE) 
				{
					_tprintf( _T("Retrieving volume name for %s failed, there is no volume mounted as this drive\n"), szDriveLetter );
					return (bFlag);
				}

				_tprintf( _T("Volume name of %s is %s\n"), szDriveLetter, szVolumeName );
			}
			else
			{
				// volume name appear to be specified directly
				_tcsncpy( szVolumeName, argv[1], NUMBER_OF(szVolumeName) );
				// make sure the string is NULL terminated
				szVolumeName[NUMBER_OF(szVolumeName)-1] = '\0';
			}

			// Make sure the volume name is terminated in a slash
			size_t iVolumeNameLength = _tcslen(szVolumeName);
			if (iVolumeNameLength && iVolumeNameLength < NUMBER_OF(szVolumeName) && szVolumeName[iVolumeNameLength-1] != '\\')
			{
				_tcscat(szVolumeName, _T("\\"));
			}

			// make sure the mount point exist
			if (_tmkdir(szMountPoint))
			{
				if (errno!=EEXIST)
				{
					TCHAR szDrive[_MAX_DRIVE + 1];
					_tsplitpath( szMountPoint, szDrive, NULL, NULL, NULL);
					_tcscat(szDrive, _T("\\"));
					// check whether path is not only the drive
					if ( _tcscmp(szMountPoint, szDrive) )
					{
						_tprintf (_T("Could not create mount point %s, parent path does not exist\n"), szMountPoint);
						return -1;
					}
					else
					{
						// We attempt to mount a volume as drive letter - for the operation to succeed, the volume must not be already mounted as another drive letter
						TCHAR szDriveRoot[] = _T("C:\\");
						for (int iDriveAssigned = 0; iDriveAssigned < NUMBER_OF(g_szDriveAssignement); iDriveAssigned++)
						{
							if ( g_szDriveAssignement[iDriveAssigned] != NULL && !_tcsicmp(szVolumeName, g_szDriveAssignement[iDriveAssigned]) )
							{
								*szDriveRoot = iDriveAssigned + 'A';
								break;
							}
						}

						if (iDriveAssigned < NUMBER_OF(g_szDriveAssignement))
						{
							_tprintf (_T("Volume %s is mounted as %s\n"), szVolumeName, szDriveRoot);
							_tprintf (_T("A volume already mounted as a drive root cannot be remounted as another drive root\n"));
							exit (-1);
						}
					}
				}
			}

			// make sure the mount point is an empty directory
			_tfinddata_t fdFindData;
			TCHAR szFindBuffer[ NUMBER_OF(szMountPoint) + 3];
			_tcscpy(szFindBuffer, szMountPoint);
			_tcscat(szFindBuffer, _T("*.*"));
			intptr_t iFFData = _tfindfirst( szFindBuffer, &fdFindData );
			if (iFFData != -1)
			{
				do
				{
					if (fdFindData.name[0] != '.')
					{
						_tprintf (_T("Could not use %s as mount point, directory is not empty\n"), szMountPoint);
						_findclose(iFFData);
						return (bFlag);
					}
				} 
				while (!_tfindnext(iFFData, &fdFindData));
			}
			_findclose(iFFData);

			// Try to mount the volume using the specified mount point
			bFlag = (*g_lpfnSetVolumeMountPoint)(szMountPoint, szVolumeName );

			if (!bFlag) 
			{
				DWORD dwErr = GetLastError();
				// Currently, reparse points are supported only on NTFS volumes. 
				_tprintf (_T("Attempt to mount %s at %s failed, error code %d\n"), szVolumeName, szMountPoint, dwErr);
				if (dwErr==ERROR_INVALID_FUNCTION)
				{
					_tprintf (_T("Mount points are currently supported only on NTFS5 drives\n"));
				}
			}
			else
			{
				_tprintf (_T("Volume successfully mounted at %s\n"), szMountPoint);
			}
		}
		else
		{
			bFlag = (*g_lpfnGetVolumeNameForVolumeMountPoint)(
				szMountPoint, // input volume mount point or directory
				szVolumeName, // output volume name buffer
				NUMBER_OF(szVolumeName) // size of volume name buffer
				);

			if (bFlag != TRUE) 
			{
				_tprintf( _T("Could not retrieve volume name mounted at %s\n"), szMountPoint );
			}
			else
			{
				_tprintf( _T("Unmounting volume %s\n"), szVolumeName );
				// Try to unmount the volume from the specified mount point
				bFlag = (*g_lpfnDeleteVolumeMountPoint)( szMountPoint );

				if (!bFlag) 
				{
					DWORD dwErr = GetLastError();
					_tprintf (_T("Attempt to unmount volume at %s failed, error code %d\n"), szMountPoint, dwErr);
				}
				else
				{
					_tprintf (_T("Volume successfully unmounted from %s\n"), szMountPoint);
					// delete the mount path
					_trmdir(szMountPoint);
				}
			}
		}
	}

	return (bFlag);
}